//----------------------------------------------------------------------------
// NAME: BigScreenshot
// VERSION: 0.3
// DESCRIPTION: Save the current screen at 2x or 3x its size. Adds itself to the Plugins menu.
// DEVELOPER: Patrick Murris
// WEBSITE: http://www.alpix.com/3d/worldwin
//----------------------------------------------------------------------------
// 0.3		Nov 21, 2006	Added 3x tiled rendering for terrain only
// 0.2		Nov 13, 2006	Fixed viewport update for Camera
// 0.1		Nov 13, 2006	Check for hardware caps and down size if needed
// 0.1b1	Nov 12, 2006	First version... 2x only, fails above some size
//----------------------------------------------------------------------------
// Known issues
// * Tiled rendering (3x) uses disk temp .png that slow down the process a lot and freeze WW
//----------------------------------------------------------------------------

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using System.Windows.Forms;
using WorldWind.Renderable;
using WorldWind.Camera;
using WorldWind;
using System.IO;
using System.Drawing;
using System.Drawing.Imaging;
using System.Globalization;
using System;

namespace Murris.Plugins
{
	/// <summary>
	/// The plugin (main class)
	/// </summary>
	public class BigScreenshot : WorldWind.PluginEngine.Plugin 
	{
		public static string pluginName = "Big Screenshot";	// Name displayed in Plugins menu
		MenuItem menuItem2x;
		MenuItem menuItem3x;
		BigScreenshotLayer layer;
		
		// Our save parameters
		public float sizeFactor;		// Viewport size multiplier (2 or 3)
		public string saveFilePath = "";	// Image file path and name
		public ImageFileFormat saveFileFormat = ImageFileFormat.Jpg;
		public ImageFormat saveFileFormat2 = ImageFormat.Jpeg;
		public string errorMsg = "";		// Last render exception

		/// <summary>
		/// Plugin entry point - All plugins must implement this function
		/// </summary>
		public override void Load() 
		{
			// Add our menu in Plugins
			menuItem2x = new MenuItem( pluginName + " 2x" );
			menuItem2x.Click += new EventHandler( menuItemClicked );
			ParentApplication.PluginsMenu.MenuItems.Add( menuItem2x );
			menuItem3x = new MenuItem( pluginName + " 3x" );
			menuItem3x.Click += new EventHandler( menuItemClicked );
			ParentApplication.PluginsMenu.MenuItems.Add( menuItem3x );

			// Add our renderable layer
			layer = new BigScreenshotLayer(pluginName, Application.WorldWindow, this);
			Application.WorldWindow.CurrentWorld.RenderableObjects.Add(layer);


		}

		/// <summary>
		/// Unloads our plugin
		/// </summary>
		public override void Unload() 
		{
			ParentApplication.PluginsMenu.MenuItems.Remove( menuItem2x );
			ParentApplication.PluginsMenu.MenuItems.Remove( menuItem3x );
			if(layer != null)
			{
				Application.WorldWindow.CurrentWorld.RenderableObjects.Remove(layer);
				layer.Dispose();
				layer = null;
			}

		}

		/// <summary>
		/// menuItem2x click handler
		/// </summary>
		private void menuItemClicked(object sender, System.EventArgs e) {
			MenuItem menuItem = (MenuItem)sender;
			// Screenshot dialog here
			string filePath = FileDialog();
			if(filePath != "") {
				// Check file format from extension
				FileInfo saveFileInfo = new FileInfo(filePath);
				string ext = saveFileInfo.Extension.Replace(".","");
				try
				{
					saveFileFormat = (ImageFileFormat) Enum.Parse(typeof(ImageFileFormat), ext, true );
					if(saveFileFormat == ImageFileFormat.Bmp) saveFileFormat2 = ImageFormat.Bmp;
					if(saveFileFormat == ImageFileFormat.Jpg) saveFileFormat2 = ImageFormat.Jpeg;
					if(saveFileFormat == ImageFileFormat.Png) saveFileFormat2 = ImageFormat.Png;
					if(menuItem == menuItem2x) sizeFactor = 2.0f;
					if(menuItem == menuItem3x) sizeFactor = 3.0f;
					saveFilePath = filePath;
				}
				catch(ArgumentException)
				{
					throw new ApplicationException("Unknown file type/file extension for file '"+saveFilePath+"'.  Unable to save.");
				}

				// Render and save
				try
				{
					//SaveScreenshot(); // NOT USED - USE ASYNCH LAYER RENDER
				}
				catch(Exception ex)
				{
					MessageBox.Show(ex.ToString(), "Error", MessageBoxButtons.OK, MessageBoxIcon.Error );
				}
			}

		}

		private string FileDialog()
		{
			WorldWindow worldWindow = ParentApplication.WorldWindow;
			SaveFileDialog dlg = new SaveFileDialog();
			dlg.Title = "Save screenshot as";
			dlg.RestoreDirectory = true;
			dlg.Filter = "Portable Network Graphics (*.png)|*.png|JPEG (*.jpg)|*.jpg|Windows Bitmap (*.bmp)|*.bmp|Targa (*.tga)|*.tga";
			dlg.FileName = string.Format( CultureInfo.InvariantCulture,
				"Screen_{0:f5}{1}_{2:f5}{3}",
				Math.Abs(worldWindow.DrawArgs.WorldCamera.Longitude.Degrees),
				worldWindow.DrawArgs.WorldCamera.Longitude > Angle.Zero ? "E" : "W",
				Math.Abs( worldWindow.DrawArgs.WorldCamera.Latitude.Degrees ),
				worldWindow.DrawArgs.WorldCamera.Latitude > Angle.Zero ? "N" : "S" );

			if (dlg.ShowDialog() != DialogResult.OK)
				return "";

			// SaveFileDialog fails to add extension when the filename contains dots.
			string ext = "*" + Path.GetExtension(dlg.FileName).ToLower(CultureInfo.InvariantCulture);
			if(dlg.Filter.IndexOf(ext)<0)
			{
				// Extension missing, add it
				ext = dlg.Filter.Split('|')[ dlg.FilterIndex * 2 - 1 ].Substring(1);
				dlg.FileName += ext;
			}

			return dlg.FileName;
		}



		/// <summary>
		/// Render and save current view with a bigger size (NOT USED)
		/// Does not handle depth buffer!
		/// </summary>
		public void SaveScreenshot() {
			DrawArgs drawArgs = ParentApplication.WorldWindow.DrawArgs;
			Device device = drawArgs.device;

			// New size = current * sizeFactor
			int width = (int)(device.Viewport.Width * sizeFactor);
			int height = (int)(device.Viewport.Height * sizeFactor);

			// Render to surface setup
			Texture renderTexture = null;
			Surface renderSurface = null;
			RenderToSurface rts = null;	// Helper
			rts = new RenderToSurface(device, width, height, 
				Format.X8R8G8B8, true, DepthFormat.D16);
			renderTexture = new Texture(device, width, height, 
    				1, Usage.RenderTarget, Format.X8R8G8B8, Pool.Default);
			renderSurface = renderTexture.GetSurfaceLevel(0);

			// Viewport
			Viewport viewport = new Viewport();
			viewport.Width = width;
			viewport.Height = height;

			// Render world
			drawArgs.BeginRender();
			drawArgs.WorldCamera.Update(device);
			rts.BeginScene(renderSurface, viewport);
			device.Clear(ClearFlags.Target | ClearFlags.ZBuffer, Color.Black, 1.0f,0);
			ParentApplication.WorldWindow.CurrentWorld.Render(drawArgs);
			rts.EndScene(Filter.None);

			// Save render surface to file
			SurfaceLoader.Save(saveFilePath, saveFileFormat, renderSurface);

			// Cleanup
			renderTexture.Dispose();
			renderTexture = null;
			rts.Dispose();
			rts = null;
		}

	}

	public class BigScreenshotLayer : RenderableObject
	{
		BigScreenshot parent;		// Ref to parent plugin
		WorldWindow worldWin;		// Current WorldWindow
		bool isRendering = false; 	// Flag to avoid recusion

		/// <summary>
		/// Constructor
		/// </summary>
		public BigScreenshotLayer(string LayerName, WorldWindow worldWindow, BigScreenshot caller) : base(LayerName)
		{
			this.RenderPriority = RenderPriority.Icons;
			this.worldWin = worldWindow;
			this.parent = caller;
		}
		/// <summary>
		/// This is where we do our rendering
		/// Called from UI thread = UI code safe in this function
		/// </summary>
		public override void Render(DrawArgs drawArgs)
		{
			// Avoid recursion when we render from inside render
			if(isRendering) return;

			// Check if we have something to save to
			if(parent.saveFilePath == "") return;

			isRendering = true;

			try
			{
				if(parent.sizeFactor <= 2) 
				{
					SaveScreenshot(drawArgs); // Using bigger viewport
				} else {
					SaveScreenshot2(drawArgs, 3); // Using tiles 3x3
				}

			}
			catch(Exception ex)
			{
				parent.errorMsg = ex.ToString();
				MessageBox.Show(ex.ToString(), "Error", MessageBoxButtons.OK, MessageBoxIcon.Error );
			}
			finally
			{
				parent.saveFilePath = "";
				isRendering = false;
			}
		}

		// Save screenshot using a bigger viewport
		public void SaveScreenshot(DrawArgs drawArgs) {
			Device device = drawArgs.device;

			// New size = current * sizeFactor
			int width = (int)(device.Viewport.Width * parent.sizeFactor);
			int height = (int)(device.Viewport.Height * parent.sizeFactor);

			// Check for max supported size for textures and rescale if necessary
			int maxWidth = device.DeviceCaps.MaxTextureWidth;
			int maxHeight = device.DeviceCaps.MaxTextureHeight;
			if(width > maxWidth) {
				float newFactor = (float)maxWidth / (float)device.Viewport.Width;
				width = maxWidth;
				height = (int)(device.Viewport.Height * newFactor);
			}
			if(height > maxHeight) {
				float newFactor = (float)maxHeight / (float)device.Viewport.Height;
				width = (int)(device.Viewport.Width * newFactor);
				height = maxHeight;
			}

			// Render to surface setup
			Texture renderTexture = null;
			Surface renderSurface = null;
			renderTexture = new Texture(device, width, height, 
    				1, Usage.RenderTarget, Format.X8R8G8B8, Pool.Default);
			renderSurface = renderTexture.GetSurfaceLevel(0);

			// Create our depth buffer
			Surface origDepth = device.DepthStencilSurface;  // Save current
			Surface depthSurface = device.CreateDepthStencilSurface(width, height, 
				DepthFormat.D16, MultiSampleType.None, 0, true);
			device.DepthStencilSurface = depthSurface;

			// Bigger Viewport
			Viewport origVp = device.Viewport;
			Viewport viewport = new Viewport();
			viewport.Width = width;
			viewport.Height = height;
			device.Viewport = viewport;
			drawArgs.WorldCamera.Update(device);	// Update Camera viewport

			// Use our surface as backbuffer
			Surface backBuffer = device.GetRenderTarget(0);
			device.SetRenderTarget(0, renderSurface);

			// Render world
			device.Clear(ClearFlags.Target | ClearFlags.ZBuffer, Color.Black, 1.0f,0);
			worldWin.CurrentWorld.Render(drawArgs);

			// Restore our original back buffer, depth buffer and viewport
			device.SetRenderTarget(0, backBuffer);
			device.DepthStencilSurface = origDepth;
			device.Viewport = origVp;

			// Save render surface to file
			SurfaceLoader.Save(parent.saveFilePath, parent.saveFileFormat, renderSurface);

			// Cleanup
			renderTexture.Dispose();
			renderTexture = null;
			depthSurface.Dispose();
			depthSurface = null;
		}

		// Save screenshot using tiles
		// from http://www.mvps.org/directx/articles/tilerender/index.htm
		public void SaveScreenshot2(DrawArgs drawArgs, int numTiles) {
			Device device = drawArgs.device;
			Color backgroundColor = SkyColor(drawArgs);

			// Ouput size = current screen size * number of tiles
			int width = device.Viewport.Width * numTiles;
			int height = device.Viewport.Height * numTiles;

			// Create the final output bitmap
            		Bitmap outBitmap = new Bitmap(width, height, System.Drawing.Imaging.PixelFormat.Format32bppArgb);
            		Graphics outGraphics = Graphics.FromImage(outBitmap);

			// Render to surface setup
			Texture renderTexture = null;
			Surface renderSurface = null;
			renderTexture = new Texture(device, device.Viewport.Width, device.Viewport.Height, 
    				1, Usage.RenderTarget, Format.X8R8G8B8, Pool.Default);
			renderSurface = renderTexture.GetSurfaceLevel(0);
			//Image renderImage = Image.FromHbitmap(renderSurface.GetGraphics().GetHdc());

			// Use our surface as backbuffer
			Surface backBuffer = device.GetRenderTarget(0);
			device.SetRenderTarget(0, renderSurface);

			// Save original projection and get a new copy
			Matrix origProjection = drawArgs.device.Transform.Projection;
			Matrix newProjection = drawArgs.device.Transform.Projection;

			// Scale new projection on x and y effectively 'zooming' in
            		newProjection.M11 *= numTiles;
            		newProjection.M22 *= numTiles;

			// --- Loop through tiles
			for(int x = 0; x < numTiles; x++) {
				for(int y = 0; y < numTiles; y++) {

					// Offset projection according to current tile being rendered
					newProjection.M31 = -((numTiles-1) - x * 2.0f);	// x axis
					newProjection.M32 = (numTiles-1) - y * 2.0f;	// y axis

					// Set new projection
					device.Transform.Projection = newProjection;

					// Render world
					device.Clear(ClearFlags.Target | ClearFlags.ZBuffer, backgroundColor, 1.0f,0);
					//worldWin.CurrentWorld.Render(drawArgs);
					// Render the active terrain layers only
					RenderMap( drawArgs, worldWin.CurrentWorld.RenderableObjects );

					// Copy tile to output bitmap (not working yet...)
            				//outGraphics.DrawImage((Image)renderSurface, new Point(device.Viewport.Width * x, device.Viewport.Height * y));
					
					// Save tile to disk
					string ext = Path.GetExtension(parent.saveFilePath);
					string tileFileName = parent.saveFilePath.Replace(ext, "_" + x.ToString() + y.ToString() + ".png");
					SurfaceLoader.Save(tileFileName, ImageFileFormat.Png, renderSurface);

					// Now load back the tile into a bitmap (it sucks i know...)
					Bitmap tileBitmap = new Bitmap(tileFileName);
					
					// Copy tile to final output image
            				outGraphics.DrawImage(tileBitmap, new Point(device.Viewport.Width * x, device.Viewport.Height * y));

					// Dispose of tile bitmap
					tileBitmap.Dispose();
					File.Delete(tileFileName);
					
				}
			}
			// --- End tile loop

			// Restore our original back buffer and projection
			device.SetRenderTarget(0, backBuffer);
			device.Transform.Projection = origProjection;

			// Save output bitmap to file
			outBitmap.Save(parent.saveFilePath, parent.saveFileFormat2);

			// Cleanup
			outBitmap.Dispose();
			outGraphics.Dispose();
			renderTexture.Dispose();
			renderTexture = null;
		}

		/// <summary>
		/// Render the active map imagery layers only
		/// From Bjorn Reppen MiniGlobe.cs
		/// </summary>
		void RenderMap(DrawArgs drawArgs, RenderableObject ro)
		{
			if(!ro.IsOn)
				return;

			if(ro is RenderableObjectList)
			{
				RenderableObjectList rol = (RenderableObjectList) ro;
				foreach(RenderableObject child in rol.ChildObjects)
					RenderMap(drawArgs, child);
				return;
			}

			if (ro is QuadTileSet)
				// Render only main imagery layers
				ro.Render(drawArgs);
		}

		// Return WW default sky color depending on camera altitude
		public Color SkyColor(DrawArgs drawArgs) {
			// Render the sky according to view - example, close to earth, render sky blue, render space as black
			System.Drawing.Color backgroundColor = System.Drawing.Color.Black;

			if(drawArgs.WorldCamera.Altitude < 1000000f)
			{
				float percent = 1 - (float)(drawArgs.WorldCamera.Altitude / 1000000);
				if(percent > 1.0f)
					percent = 1.0f;
				else if(percent < 0.0f)
					percent = 0.0f;

				backgroundColor = System.Drawing.Color.FromArgb(
					(int)(World.Settings.SkyColor.R*percent),
					(int)(World.Settings.SkyColor.G*percent),
					(int)(World.Settings.SkyColor.B*percent));
			}
			return backgroundColor;
		}


		public override void Initialize(DrawArgs drawArgs)
		{
			isInitialized = true;
		}

		public override void Update(DrawArgs drawArgs)
		{
		}

		public override void Dispose()
		{
			isInitialized = false;
		}
		
		public override bool PerformSelectionAction(DrawArgs drawArgs)
		{
			return false;
		}

	}

}
